use log::debug;
use sqlx::{migrate::MigrateDatabase, Sqlite, SqlitePool};

use crate::sheet::Sheet;

pub struct Database {
    connection: SqlitePool,
}

impl Database {
    pub async fn setup(database_path: &str) -> sqlx::Result<Self> {
        let connection = Database::create_connection(database_path).await?;
        sqlx::migrate!("./db-migrations").run(&connection).await?;
        Ok(Database { connection })
    }

    async fn create_connection(database_path: &str) -> sqlx::Result<SqlitePool> {
        if !Sqlite::database_exists(database_path)
            .await
            .unwrap_or(false)
        {
            match Sqlite::create_database(database_path).await {
                Ok(_) => debug!("Database created"),
                Err(e) => {
                    debug!("Error creating database: {}", e);
                    return Err(e);
                }
            }
        }

        let connection = SqlitePool::connect(database_path).await?;
        debug!("Connected to database");
        Ok(connection)
    }

    pub async fn insert_sheet(&self, sheet: Sheet) -> sqlx::Result<()> {
        sqlx::query(
            "
                INSERT INTO sheets (name, path, file_size, file_hash)
                VALUES ($1, $2, $3, $4)
            ",
        )
        .bind(sheet.name)
        .bind(sheet.path)
        .bind(sheet.file_size)
        .bind(sheet.file_hash)
        .execute(&self.connection)
        .await
        .map(|_| ())
    }

    pub async fn update_sheet_path(&self, sheet: Sheet) -> sqlx::Result<()> {
        sqlx::query("UPDATE sheets SET path = $1 WHERE id = $2")
            .bind(sheet.path)
            .bind(sheet.id)
            .execute(&self.connection)
            .await
            .map(|_| ())
    }

    pub async fn fetch_all_sheets(&self) -> sqlx::Result<Vec<Sheet>> {
        sqlx::query_as::<_, Sheet>("SELECT * FROM sheets")
            .fetch_all(&self.connection)
            .await
    }
}
