mod database;
mod mcdu;
mod sheet;
mod sheet_listing;

use std::{
    path::{Path, PathBuf},
    process,
};

use clap::Parser;
use database::Database;
use env_logger::Env;
use gtk::prelude::*;
use log::{debug, error};
use mcdu::McduModel;
use relm4::prelude::*;
use sheet::Sheet;
use sheet_listing::{SheetListingInput, SheetListingModel};
use walkdir::WalkDir;

struct AppModel {
    mcdu: Controller<McduModel>,
    sheet_listing: Controller<SheetListingModel>,
}

#[derive(Debug)]
enum AppInput {
    SearchStarted(String),
    SheetPressed(PathBuf),
}

#[relm4::component]
impl SimpleComponent for AppModel {
    type Input = AppInput;
    type Output = ();
    type Init = PathBuf;

    view! {
        #[root]
        gtk::Window{
            set_default_width: 300,
            set_default_height: 300,
            set_title: Some("Play music!!!"),
            set_maximized: true,
            gtk::Box {
                set_orientation: gtk::Orientation::Horizontal,
                gtk::Box {
                    set_orientation: gtk::Orientation::Vertical,
                    set_hexpand: true,
                    gtk::ScrolledWindow {
                        model.sheet_listing.widget(),
                        set_vexpand: true,
                        set_hexpand: true,
                        // set_hscrollbar_policy: PolicyType::Never,
                    },
                },
                model.mcdu.widget(),
            }
        }
    }

    fn init(
        path: Self::Init,
        window: &Self::Root,
        sender: ComponentSender<Self>,
    ) -> relm4::ComponentParts<Self> {
        relm4_icons::initialize_icons();

        let mcdu = McduModel::builder()
            .launch(())
            .forward(sender.input_sender(), |response| match response {
                mcdu::McduOutput::SearchStarted(query) => AppInput::SearchStarted(query),
            });

        let sheet_listing =
            SheetListingModel::builder()
                .launch(path)
                .forward(sender.input_sender(), |response| match response {
                    sheet_listing::SheetPressedMessage::SheetPressed(path) => {
                        AppInput::SheetPressed(path)
                    }
                });

        let model = AppModel {
            mcdu,
            sheet_listing,
        };

        let widgets = view_output!();

        ComponentParts { model, widgets }
    }

    fn update(&mut self, message: Self::Input, _sender: ComponentSender<Self>) {
        match message {
            AppInput::SheetPressed(sheet) => opener::open(sheet).unwrap(),
            AppInput::SearchStarted(query) => self.sheet_listing.emit(SheetListingInput { query }),
        }
    }
}

#[derive(Parser)]
#[command(author, version, about)]
struct Cli {
    directory: PathBuf,
}

#[tokio::main]
async fn main() {
    env_logger::Builder::from_env(Env::default().default_filter_or("debug")).init();
    let cli = Cli::parse();
    if !cli.directory.is_dir() {
        error!("Sheet folder path is no dir or does not exist");
        process::exit(1);
    }

    let database = Database::setup("./testdb.sqlite").await.unwrap();
    // database.insert_sheet(Sheet::new_debug()).await.unwrap();
    let sheets = database.fetch_all_sheets().await.unwrap();

    debug!("Validating sheets from database...");
    let validation_result = validate_sheet_files(sheets, &cli.directory);
    debug!("{}", validation_result.get_stats());
    for updated in validation_result.updated_sheets {
        database.update_sheet_path(updated).await.unwrap();
    }

    let app = RelmApp::new("de.frajul.sheet-organizer");
    // Pass empty command line args to allow my own parsing
    app.with_args(Vec::new()).run::<AppModel>(cli.directory);
}

struct FileValidationResult {
    validated_sheets: Vec<Sheet>,
    invalidated_sheets: Vec<Sheet>,
    updated_sheets: Vec<Sheet>,

    unassigned_files: Vec<PathBuf>,
}

impl FileValidationResult {
    fn get_stats(&self) -> String {
        format!("Validated sheets: {}\nInvalidated sheets: {}\nUpdated sheets: {}\nUnassigned files: {}", self.validated_sheets.len(), self.invalidated_sheets.len(), self.updated_sheets.len(), self.unassigned_files.len())
    }
}

fn validate_sheet_files(sheets: Vec<Sheet>, dir: impl AsRef<Path>) -> FileValidationResult {
    let (validated_sheets, mut invalidated_sheets): (Vec<_>, Vec<_>) = sheets
        .into_iter()
        .partition(|sheet| sheet.validate_path(&sheet.path).unwrap_or(false));

    let mut updated_sheets = Vec::new();
    let mut unassigned_files = Vec::new();

    for pdf_file in WalkDir::new(dir)
        .into_iter()
        .filter_map(|e| e.ok())
        .filter(|file| file.file_type().is_file())
        .map(|file| file.into_path())
        .filter(|path| {
            path.extension()
                .map(|s| s.to_string_lossy().to_ascii_lowercase() == "pdf")
                .unwrap_or(false)
        })
    {
        if let Some((i, _)) = invalidated_sheets
            .iter()
            .enumerate()
            .find(|(_, sheet)| sheet.validate_path(&pdf_file).unwrap_or(false))
        {
            let mut sheet = invalidated_sheets.remove(i);
            sheet.path = pdf_file.to_str().unwrap().to_string();
            updated_sheets.push(sheet);
        } else {
            unassigned_files.push(pdf_file);
        }
    }

    FileValidationResult {
        validated_sheets,
        invalidated_sheets,
        updated_sheets,
        unassigned_files,
    }
}
